part of json_object;

class _KeyValuePair {
  var key;
  var value;
  
  _KeyValuePair(this.key,this.value);
}

  
/** Optional [key] is required in case the object passed in came from a map
 *  we need the original map's key to be able to re-constitute the map
 *  later when the future completes. 
 */ 
Future objectToSerializable(Object object, [key=null, depth=0]) {
  print("objectToSerializable: $object, $key, $depth");
  Completer completer = new Completer();
    
  if (object is num || 
      object is bool || 
      object is String || 
      object == null) {
    _serializeNative(object, completer, key, depth);
  }
  else if (object is Map) {
    _serializeMap(object, completer, key, depth);
  }
  else if (object is List) {
    _serializeList(object, completer, key, depth);
  }
  else {
    var instanceMirror = mirrors.reflect(object);
    _serializeObject(instanceMirror, completer, key, depth);
    // all other processing of regular classes    
  }
   
  return completer.future;
}

void _serializeNative(Object object, Completer completer, key, depth) {
   print("native: $object");
  // "native" object types - just complete with that type
  _complete(completer,object,key, depth);
}

void _serializeMap(Map object, Completer completer, key, depth) {
  print("map: $object");
  
  // convert the map into a serialized map
  // each value in the map may itself be a complex object or a "native" type.
  // so we need to test for this  
  Map<String, Future> mapItemsToComplete = new Map<String, Future>();
  object.forEach((key,value) {
    mapItemsToComplete[key] = objectToSerializable(value,key); 
  });
  
  var onAllItemsComplete = (List keyValuePairs) {
    // at this point (via the Future.wait callback)
    // all items in the map are complete.
    // but how to match the items back to the keys?
    var mapResult = new Map();
    keyValuePairs.forEach((kv) => mapResult[kv.key] = kv.value);
    _complete(completer,mapResult,key, depth);
  };
  
  var onItemsError = (error) => completer.completeError(error);
  
  Future.wait(mapItemsToComplete.values)
    .then(onAllItemsComplete, onError:onItemsError);
}

void _serializeList(List object, Completer completer, key, depth) {
   print("list: $object");
  
  // each item in the list will be an object to serialize.
  List<Future> listItemsToComplete = new List<Future>();
  object.forEach((item) {
    listItemsToComplete.add(objectToSerializable(item));
  });
  
  var onAllItemsComplete = (items) => _complete(completer,items,key, depth);
  var onItemsError = (error) => completer.completeError(error);
  
  Future.wait(listItemsToComplete)
    .then(onAllItemsComplete, onError:onItemsError);
}

void _serializeObject(mirrors.InstanceMirror instanceMirror, Completer completer, key, depth) {
  print("object: $instanceMirror, $key, $depth");
  var classMirror = instanceMirror.type;
  
  var resultMap = new Map();
  //var futuresList = new List<Future>();
  
    // for each getter:
   classMirror.getters.forEach((getterKey, getter) {
      if (!getter.isPrivate && !getter.isStatic) {
        var futureField = instanceMirror.getField(getterKey);
        resultMap[getterKey] = futureField;
      }
    });
    
    // for each field
    classMirror.variables.forEach((varKey, variable) {
      if (!variable.isPrivate && !variable.isStatic) {
        var futureField = instanceMirror.getField(varKey);
        resultMap[varKey] = futureField;
      }
      
    });  
  
    _futureMapWait(resultMap).then((Map<String,Object> result) {
      print("Future result: $result");
      
      result.forEach((k,v) {
        if (v is mirrors.InstanceMirror) {
          result[k] = v.reflectee;
        }
      });
      
      objectToSerializable(result, key, depth+1).then((res) {
        _complete(completer, res, null, depth);
      });
      
      
      
    });      
}

void _complete(completer, object, key, depth) {
  print("Completing: $object, $key, $depth");
  if (key != null) {      
    completer.complete(new _KeyValuePair(key,object)); // complete, because we can't reflect any deeper
  }
  else {
    completer.complete(object); // complete, because we can't reflect any deeper
  }
}


/**
 * A modified, recursive implementation of 
 * factory _FutureImpl.wait(Iterable<Future> futures)
 * to wait() for a map whose values may contain futures.  
 */
Future<Map<String,Object>> _futureMapWait(Map<String, Object> futures) {
  Map<String,Object> result = new Map<String,Object>();
   
  if (futures.isEmpty) {
    return new Future<Map<String,Object>>.immediate(result);
  }

  Completer completer = new Completer<Map<String,Object>>();
  int remaining = futures.length;

  // As each future completes, put its value into the corresponding
  // position in the list of values.
  int i = 0;
  bool completed = false;
  
  for (var key in futures.keys) {
    int pos = i++;
    // todo: possibly recurse here?
    if (futures[key] is Future) {
      (futures[key] as Future).then((Object value) {
        result[key] = value;
        if (--remaining == 0) {
          completer.complete(result);
        }
      }).catchError((error) {
        if (!completed) completer.completeError(error.error, error.stackTrace);
        completed = true;
      });
    }
    else if (futures[key] is Map) { // it may be a map containing other futures
      print("recursing");
      _futureMapWait(futures[key]).then((Map<String, Object> values) {
        result[key] = values;
        if (--remaining == 0) {
          completer.complete(result);
        }
      });
    }
    else { // list?
      // not a future, or map that might be containing futures
      result[key] = futures[key];
      if (--remaining == 0) {
        completer.complete(result);
      }
    }
  }

  return completer.future;
}